package pps.core.common.utils;

import com.google.common.base.Joiner;
import com.google.common.base.Strings;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.http.HttpServletRequest;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

/**
 * IP地址
 *
 * @author lixueyan
 */
public class IpUtils {
    private static Logger logger = LoggerFactory.getLogger(IpUtils.class);

    /**
     * 获取IP地址
     * <p>
     * 使用Nginx等反向代理软件， 则不能通过request.getRemoteAddr()获取IP地址
     * 如果使用了多级反向代理的话，X-Forwarded-For的值并不止一个，而是一串IP地址，X-Forwarded-For中第一个非unknown的有效IP字符串，则为真实IP地址
     */
    public static String getIpAddr(HttpServletRequest request) {
        String unknown = "unknown";
        String ip = null;
        try {
            ip = request.getHeader("x-forwarded-for");
            if (StringUtils.isEmpty(ip) || unknown.equalsIgnoreCase(ip)) {
                ip = request.getHeader("Proxy-Client-IP");
            }
            if (StringUtils.isEmpty(ip) || ip.length() == 0 || unknown.equalsIgnoreCase(ip)) {
                ip = request.getHeader("WL-Proxy-Client-IP");
            }
            if (StringUtils.isEmpty(ip) || unknown.equalsIgnoreCase(ip)) {
                ip = request.getHeader("HTTP_CLIENT_IP");
            }
            if (StringUtils.isEmpty(ip) || unknown.equalsIgnoreCase(ip)) {
                ip = request.getHeader("HTTP_X_FORWARDED_FOR");
            }
            if (StringUtils.isEmpty(ip) || unknown.equalsIgnoreCase(ip)) {
                ip = request.getRemoteAddr();
            }
        } catch (Exception e) {
            logger.error("IPUtils ERROR ", e);
        }

        return ip;
    }

    private volatile static String ips;

    public synchronized static String getHostIPs() {
        return getHostIPs(false);
    }

    public static String getHostIPs(boolean refresh) {
        if(!refresh && !Strings.isNullOrEmpty(ips)){
            return ips;
        }
        
        List<String> list = new ArrayList<String>();
        try {
            Enumeration<NetworkInterface> en = NetworkInterface.getNetworkInterfaces();
            while (en.hasMoreElements()) {
                NetworkInterface nint = en.nextElement();
                if (nint.isLoopback()) {
                    continue;
                }
                if (nint.isPointToPoint()) {
                    continue;
                }
                if (nint.isVirtual()) {
                    continue;
                }
                if(!nint.isUp()){
                    continue;
                }
                List<InterfaceAddress> interfaceAddresses = nint.getInterfaceAddresses();
                if (interfaceAddresses == null) {
                    continue;
                }
                interfaceAddresses.forEach(item -> {
                    String addr = item.getAddress().getHostAddress();
                    if (!Strings.isNullOrEmpty(addr)) {
                        list.add(addr.trim());
                    }
                });
            }
        } catch (java.net.SocketException e) {
            // fine, let's take is as signal of not having any interfaces
        }
        list.sort((x, y) -> {
            return x.compareTo(y);
        });
        ips = Joiner.on(",").join(list);
        return ips;
    }
}
