package pps.core.common.utils;

import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import redis.clients.jedis.*;
import xstartup.base.XServiceContext;
import xstartup.base.XStartup;
import xstartup.base.exception.XServiceException;
import xstartup.base.util.XConvertUtils;
import xstartup.base.util.XEncryptUtils;
import xstartup.base.util.XStringUtils;
import xstartup.cache.redis.XRedisCacheProvider;
import xstartup.cache.redis_cluster.XRedisClusterCacheProvider;
import xstartup.config.XConfProperties;
import xstartup.config.XConfPropertiesHelper;
import xstartup.error.XError;

import java.time.Duration;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;

/**
* @Description: Jedis工具类
* @author xcd
* @date 2022/10/27
* @version V1.0
*/
public class JedisHelper {

    public static void consumeDefalutJedis(Consumer<Jedis> consumer) {
        consumeJedis("","","", consumer);
    }

    public static void consumeJedis(String product, String scope, String module, Consumer<Jedis> consumer) {
        JedisPool jedisPool = JedisPoolTool.getJedisPoolWrap(product, scope, module).jedisPool;
        Jedis jedis = jedisPool.getResource();
        try{
            consumer.accept(jedis);
        }finally {
            if(jedis.isBroken()){
                jedisPool.returnBrokenResource(jedis);
            }else{
                jedisPool.returnResource(jedis);
            }
        }
    }


    public static void consumeDefalutJedisCluster(Consumer<JedisCluster> consumer) {
        consumeJedisCluster("","","", consumer);
    }

    public static void consumeJedisCluster(String product, String scope, String module, Consumer<JedisCluster> consumer) {
        JedisCluster resource = JedisPoolTool.getJedisClusterPoolWrap(product, scope, module).getResource();
        consumer.accept(resource);
    }

    static class JedisPoolTool {
        private static final ConcurrentHashMap<String, JedisPoolWrapper> jedisPoolMapByScope = new ConcurrentHashMap();
        private static final ConcurrentHashMap<String, JedisPoolWrapper> jedisPoolMapByHostDB = new ConcurrentHashMap();

        private static final ConcurrentHashMap<String, JedisClusterPoolWrapper> jedisClusterPoolMapByScope = new ConcurrentHashMap();
        private static final ConcurrentHashMap<String, JedisClusterPoolWrapper> jedisClusterPoolMapByHostDB = new ConcurrentHashMap();

        JedisPoolTool() {
        }

        static JedisPoolWrapper getGlobalJedisPoolWrap() {
            return getJedisPoolWrap("", "", "");
        }

        static JedisPoolWrapper getJedisPoolWrap(String product, String scope, String module) {
            String productScopeModule = product + "." + scope + "." + module;
            JedisPoolWrapper existed = jedisPoolMapByScope.get(productScopeModule);
            if (existed != null) {
                return existed;
            }

            XServiceContext serviceContext = XServiceContext.build(XRedisCacheProvider.class.getName());
            XConfProperties xConfProperties = XConfPropertiesHelper.find(serviceContext, "x.cache", "type"
                    , new String[]{"host", "port", "auth", "db"}, productScopeModule);
            if (xConfProperties == null) {
                throw new XServiceException(XError.NotFound, "缓存配置信息未找到");
            }

            String host = xConfProperties.getProperty("host");
            String auth = xConfProperties.getProperty("auth");
            try {
                auth = XEncryptUtils.decrypt(auth);
            } catch (XServiceException var19) {
                serviceContext.getLogger().warn("redis密码解密失败:" + var19.getMessage() + ",直接使用原文:" + auth);
            }
            Integer port = xConfProperties.getInteger("port", 6379);
            Integer db = xConfProperties.getInteger("db", 0);
            if (XStringUtils.isEmpty(host)) {
                throw new XServiceException(XError.NotFound, "缓存配置信息未找到(type,host)");
            }

            synchronized (jedisPoolMapByScope) {
                //再次检查
                existed = jedisPoolMapByScope.get(productScopeModule);
                if (existed != null) {
                    return existed;
                }
                String hostPortDb = host + ":" + port + ":" + db;
                JedisPoolWrapper poolWrapperByHostPortDB = jedisPoolMapByHostDB.get(hostPortDb);
                if (poolWrapperByHostPortDB == null) {
                    XStartup xstartup = XStartup.getCurrent();
                    JedisPoolConfig poolConfig = new JedisPoolConfig();
                    poolConfig.setMaxTotal(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-total"), 10));
                    poolConfig.setMaxIdle(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-idle"), 10));
                    poolConfig.setMinIdle(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.min-idle"), 0));
                    poolConfig.setMaxWait(Duration.ofMillis((long) XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-wait-millis"), 5000)));
                    poolConfig.setTestOnBorrow(false);
                    poolConfig.setTestOnReturn(false);
                    poolConfig.setTestWhileIdle(true);
                    poolConfig.setTimeBetweenEvictionRuns(Duration.ofMillis(10000L));
                    poolConfig.setMinEvictableIdleTime(Duration.ofMillis(60000L));
                    poolConfig.setNumTestsPerEvictionRun(2);
                    JedisPool jedisPool = new JedisPool(poolConfig, host, port, 3000, auth, db);
                    poolWrapperByHostPortDB = new JedisPoolWrapper(jedisPool, db, auth);
                    jedisPoolMapByHostDB.put(hostPortDb, poolWrapperByHostPortDB);
                }
                jedisPoolMapByScope.put(productScopeModule, poolWrapperByHostPortDB);
                return poolWrapperByHostPortDB;
            }
        }

        static JedisClusterPoolWrapper getGlobalJedisClusterPoolWrap() {
            return getJedisClusterPoolWrap("", "", "");
        }
        static JedisClusterPoolWrapper getJedisClusterPoolWrap(String product, String scope, String module){
            String productScopeModule = product + "." + scope + "." + module;
            CounterBuilder.globalCounterBuilder.setFieldValue("jedisClusterPoolMapByScope", jedisClusterPoolMapByScope.size()+"");
            JedisClusterPoolWrapper existed = jedisClusterPoolMapByScope.get(productScopeModule);
            if (existed != null) {
                return existed;
            }

            XServiceContext serviceContext = XServiceContext.build(XRedisCacheProvider.class.getName());
            XConfProperties xConfProperties = XConfPropertiesHelper.find(serviceContext, "x.cache", "type"
                    , new String[]{"host", "port", "auth"}, productScopeModule);
            if (xConfProperties == null) {
                throw new XServiceException(XError.NotFound, "缓存配置信息未找到");
            }

            String type = xConfProperties.getProperty("type");
            String host = xConfProperties.getProperty("host");
            String auth = xConfProperties.getProperty("auth");
            try {
                auth = XEncryptUtils.decrypt(auth);
            } catch (XServiceException var22) {
                serviceContext.getLogger().warn("redis密码解密失败:" + var22.getMessage() + ",直接使用原文:" + auth);
            }
            String port = xConfProperties.getProperty("port");

            if (XStringUtils.isEmpty(type) || XStringUtils.isEmpty(host)) {
                throw new XServiceException(XError.NotFound.getCode(), "缓存配置信息未找到");
            }

            String[] hostArr = host.split(",");
            String[] portArr = port.split(",");
            if (hostArr.length != portArr.length) {
                throw new XServiceException(XError.InvalidParameterValue.getCode(), "配置信息不正确");
            }

            synchronized(jedisClusterPoolMapByHostDB) {

                CounterBuilder.globalCounterBuilder.addCounter("JedisCluster-begin");
                productScopeModule = product + "." + scope + "." + module;
                JedisClusterPoolWrapper existed2 = jedisClusterPoolMapByScope.get(productScopeModule);
                if (existed2 != null) {
                    return existed2;
                }

                List<String> hostPortKeyList = new ArrayList();
                Set<HostAndPort> jedisClusterNode = new HashSet();
                for(int index = 0; index < hostArr.length; ++index) {
                    hostArr[index] = hostArr[index].trim();
                    portArr[index] = portArr[index].trim();
                    jedisClusterNode.add(new HostAndPort(hostArr[index], XConvertUtils.toInteger(portArr[index])));
                    hostPortKeyList.add(hostArr[index] + ":" + portArr[index]);
                }
                hostPortKeyList.sort(String::compareTo);
                String jedisPoolAddressKey = String.join(".", hostPortKeyList);

                JedisClusterPoolWrapper jedisPoolByAddress = jedisClusterPoolMapByHostDB.get(jedisPoolAddressKey);
                if( jedisPoolByAddress == null){
                    CounterBuilder.globalCounterBuilder.addCounter("JedisCluster-build-JedisClusterPoolWrapper");
                    XStartup xstartup = XStartup.getCurrent();
                    GenericObjectPoolConfig<Connection> poolConfig = new GenericObjectPoolConfig();
                    poolConfig.setMaxTotal(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-total"), 10));
                    poolConfig.setMaxIdle(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-idle"), 10));
                    poolConfig.setMinIdle(XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.min-idle"), 2));
                    poolConfig.setMaxWait(Duration.ofMillis((long)XConvertUtils.tryToInteger(xstartup.getProperty("x.cache.jedis.max-wait-millis"), 5000)));
                    poolConfig.setTestOnBorrow(false);
                    poolConfig.setTestOnReturn(false);
                    poolConfig.setTestWhileIdle(true);
                    poolConfig.setTimeBetweenEvictionRuns(Duration.ofMillis(10000L));
                    poolConfig.setMinEvictableIdleTime(Duration.ofMillis(60000L));
                    poolConfig.setNumTestsPerEvictionRun(2);
                    JedisClusterPoolWrapper jedisPoolWrapper = new JedisClusterPoolWrapper (jedisClusterNode, poolConfig, auth);
                    jedisClusterPoolMapByHostDB.put(jedisPoolAddressKey, jedisPoolWrapper);
                    jedisPoolByAddress = jedisPoolWrapper;
                }
                jedisClusterPoolMapByScope.put(productScopeModule, jedisPoolByAddress);
                CounterBuilder.globalCounterBuilder.addCounter("JedisCluster-end");
                return jedisPoolByAddress;
            }
        }
    }

    static class JedisPoolWrapper {
        final JedisPool jedisPool;
        final int db;
        final String auth;

        public JedisPoolWrapper(JedisPool jedisPool, int db, String auth) {
            this.jedisPool = jedisPool;
            this.db = db;
            this.auth = auth;
        }
    }

    static class JedisClusterPoolWrapper {
        String auth;
        private Set<HostAndPort> jedisClusterNode;
        private GenericObjectPoolConfig<Connection> poolConfig;
        private final  JedisCluster cluster ;

        public JedisClusterPoolWrapper(Set<HostAndPort> jedisClusterNode, GenericObjectPoolConfig<Connection> poolConfig, String auth) {
            this.jedisClusterNode = jedisClusterNode;
            this.poolConfig = poolConfig;
            this.auth = auth;
            cluster = new JedisCluster(this.jedisClusterNode, 6000, 5000, 10, this.auth, this.poolConfig);
        }

        public JedisCluster getResource() {
            return cluster;
        }
    }
}
