package pps.core.common.utils;

import com.google.common.base.Strings;
import org.joor.Reflect;
import xstartup.base.XContext;
import xstartup.base.util.XJsonUtils;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

public class TraceHelper {

    public static final String FORM_DATA_TRACE_BUILDER = "TRACE_LOG_BUILDER";
    public static final int TITLE_LENGTH = 30;

    public static class TraceBuilder {

        private static final String NEW_LINE = System.getProperty("line.separator");

        private final String traceId;
        private final Long traceTime = System.currentTimeMillis();
        private final List<Object> items = new ArrayList<Object>();
        private long latestTime = 0;

        public TraceBuilder(String traceId) {
            this.traceId = traceId;
        }

        public String getTraceId() {
            return traceId;
        }

        public Long getTraceTime() {
            return traceTime;
        }

        public List<Object> getItems() {
            return items;
        }

        public String buildTraceLog() {
            final StringBuilder builder = new StringBuilder();
            items.forEach(item -> {
                builder.append(item).append(NEW_LINE);
            });
            return builder.toString();
        }

        public long getLatestTime() {
            return latestTime;
        }

        public void setLatestTime(long latestTime) {
            this.latestTime = latestTime;
        }

        public void clear() {
            items.clear();
        }
    }

    //    private static final Boolean ENABLETRACE = Boolean.TRUE;
    private static final ConcurrentHashMap<String, TraceBuilder> traceBuilders = new ConcurrentHashMap<String, TraceBuilder>();
    private static TimeSpanChecker timeSpanChecker = new TimeSpanChecker(300, true);
    private static AtomicLong builtTraceTotalCount = new AtomicLong(0);
    private static AtomicLong removedTraceTotalCount = new AtomicLong(0);
    public static final String PUSH_TAG = "2022-09-01 12:21:07";

    public static long getTraceBuildersCount() {
        return traceBuilders.size();
    }


    public static long getBuiltTraceTotalCount() {
        return builtTraceTotalCount.get();
    }

    public static long getRemovedTotalCount() {
        return removedTraceTotalCount.get();
    }
//
//    public static ConcurrentLinkedQueue<String> getLastTraceIds() {
//        return lastTraceIds;
//    }
//
//    public static void setLastTraceIds(ConcurrentLinkedQueue<String> lastTraceIds) {
//        TraceHelper.lastTraceIds = lastTraceIds;
//    }

    public static int getTraceCount() {
        return traceBuilders.size();
    }

    public static TraceBuilder buildTraceBuilder(XContext context) {
        return buildTraceBuilder(context.getTraceId(), context);
    }

    public static TraceBuilder buildTraceBuilder(String traceId, XContext context) {

//        Boolean enableTraceLog = TraceLogConfig.checkEnableTraceLog(context);// TraceLogConfig.enableTraceLog != null ? TraceLogConfig.enableTraceLog.getValue(context) : null;
//        if (context != null && enableTraceLog != null && !enableTraceLog.booleanValue()) {
//            traceBuilders.clear();
//            return null;
//        }

        if (timeSpanChecker.checkNow()) {
            cleanTraceBuilder();
        }

        if (!Strings.isNullOrEmpty(traceId) && traceBuilders.contains(traceId)) {
            return traceBuilders.get(traceId);
        }

        if (Strings.isNullOrEmpty(traceId)) {
            traceId = UUID.randomUUID().toString().replace("-", "");
        }

        CounterBuilder.globalCounterBuilder.addCounter("TraceBuilder-New");
        String userName = context != null && context.getUser() != null ? (context.getUser().getName() + ",") : "无，";
        TraceBuilder builder = new TraceBuilder(traceId);

        builder.getItems().add("[" + DateUtil.getCurrentDateTimeWithMs() + "]" + StringAlignHelper.alignRight(userName + "BUILT", TITLE_LENGTH, '=') + "|" + (context == null || context.getLogTag() == null ? "" : context.getLogTag()));
        appendTraceLog(builder, "IPs", IpUtils.getHostIPs());
        appendTraceLog(builder, "Thread", Thread.currentThread().getName());
        appendTraceLog(builder, "ComponentBuildTime", ManifestComponentInfoHelper.getManifestComponentInfo("pps","pps-common").getBuildTime());

        if (context != null) {
            context.setAdditionalData(TraceHelper.FORM_DATA_TRACE_BUILDER, builder);
        }

        traceBuilders.put(traceId, builder);
        builtTraceTotalCount.incrementAndGet();
        CounterBuilder.globalCounterBuilder.addCounter("TraceHelper-builtTraceTotalCount");
        return builder;
    }

    public static TraceBuilder getTraceBuilder(String traceId) {
        try {

            if (Strings.isNullOrEmpty(traceId)) {
                return null;
            }

            if (timeSpanChecker.checkNow()) {
                cleanTraceBuilder();
            }

            return traceBuilders.get(traceId);
        } catch (Exception e) {
            return null;
        }
    }

    private static void cleanTraceBuilder() {
        synchronized (timeSpanChecker) {
            Iterator<Map.Entry<String, TraceBuilder>> it = traceBuilders.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry<String, TraceBuilder> item = it.next();
                double seconds = (System.currentTimeMillis() - item.getValue().getTraceTime()) / 1000.0;
                if (seconds > 300) {
                    it.remove();
                    removedTraceTotalCount.incrementAndGet();
                    CounterBuilder.globalCounterBuilder.addCounter("TraceHelper-removedTraceTotalCount");
                }
            }
        }
    }

    public static TraceHelper.TraceBuilder getTraceBuilder(XContext context) {
        if (context == null) {
            return null;
        }
        TraceBuilder traceBuilder = (TraceBuilder) context.getAdditionalData(FORM_DATA_TRACE_BUILDER);
        if (traceBuilder != null) {
            return traceBuilder;
        }
        return getTraceBuilder(context.getTraceId());
    }


    public static void appendTraceLog(XContext context, String title) {
        appendTraceLog(context, title, null);
    }

    public static void appendTraceLog(TraceBuilder builder, String title) {
        appendTraceLog(builder, title, null);
    }

    public static void appendTraceLog(XContext context, Throwable e) {
        appendTraceLog(context, "EXCEPTION", e.getMessage());
        appendTraceLog(context, "EXCEPTION", ExceptionHelper.getExceptionLogDetail(e));
    }

    public static void appendTraceLog(TraceBuilder builder, Throwable e) {
        appendTraceLog(builder, "EXCEPTION", e.getMessage());
        appendTraceLog(builder, "EXCEPTION", ExceptionHelper.getExceptionLogDetail(e));
    }

    public static void appendTraceLog(XContext context, String title, Object detail) {
        if (context == null) {
            return;
        }
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        if (traceBuilder == null) {
            return;
        }

        appendTraceLog(traceBuilder, title, detail);
    }

    public static void appendTraceLog(TraceBuilder traceBuilder, String title, Object detail) {
        if (traceBuilder == null) {
            return;
        }

        String detailStr = detail == null ? "" : ("====> " + detail.toString());
//        detailStr = detailStr.replace("\\\"","\"");
//        String latestTime = DateUtil.getCurrentDateTimeWithMs();
        long latestTime = System.currentTimeMillis();
        traceBuilder.getItems().add("[" + DateUtil.getCurrentDateTimeWithMs() + "]" + StringAlignHelper.alignRight(title, TITLE_LENGTH, '=') + "|" + detailStr);
        traceBuilder.setLatestTime(latestTime);
    }

    public static void appendTraceLogLineSeperator(XContext context) {
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        if (traceBuilder == null) {
            return;
        }

        appendTraceLogLineSeperator(traceBuilder);
    }

    public static void appendTraceLogLineSeperator(TraceBuilder traceBuilder) {
        if (traceBuilder == null) {
            return;
        }

        long latestTime = System.currentTimeMillis();
        traceBuilder.getItems().add("-----------------------------------------------------------------------------------------------------------------------------------");
        traceBuilder.setLatestTime(latestTime);
    }

    public static void appendTraceLogWithSpan(XContext context, String title) {
        appendTraceLogWithSpan(context, title, null);
    }

    public static void appendTraceLogWithSpan(TraceBuilder traceBuilder, String title) {
        appendTraceLogWithSpan(traceBuilder, title, null);
    }

    public static void appendTraceLogWithSpan(XContext context, String title, Object detail) {
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        appendTraceLogWithSpan(traceBuilder, title, detail);
    }

    public static void appendTraceLogWithSpan(TraceBuilder traceBuilder, String title, Object detail) {
        if (traceBuilder == null) {
            return;
        }

        long latestTime = System.currentTimeMillis();

        String detailStr = detail == null ? "" : ("====> " + detail.toString());
//        detailStr = detailStr.replace("\\\"","\"");
        detailStr += "====> USED:  " + String.format("%.3fs", (latestTime - traceBuilder.getLatestTime()) / 1000.0);
        traceBuilder.getItems().add("[" + DateUtil.getCurrentDateTimeWithMs() + "]" + StringAlignHelper.alignRight(title, TITLE_LENGTH, '=') + "|" + detailStr);
        traceBuilder.setLatestTime(latestTime);
    }

    //
//    public static void appendTraceLogCurrentTime(XContext context) {
//        appendTraceLogCurrentTime(context, false);
//    }
//
    public static void appendTraceLogUsed(XContext context) {
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        appendTraceLogUsed(traceBuilder);
    }

    public static void appendTraceLogUsed(TraceHelper.TraceBuilder traceBuilder) {
        if (traceBuilder == null) {
            return;
        }

        long latestTime = System.currentTimeMillis();
        traceBuilder.setLatestTime(latestTime);
        appendTraceLog(traceBuilder, StringAlignHelper.alignRight("USED", TITLE_LENGTH, '='), "  " + String.format("%.3fs", (latestTime - traceBuilder.getLatestTime()) / 1000.0));

    }

    public static void appendTraceLogTotalTime(XContext context) {
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        if (traceBuilder == null) {
            return;
        }

        appendTraceLogTotalTime(traceBuilder);
    }

    public static void appendTraceLogTotalTime(TraceHelper.TraceBuilder traceBuilder) {
        if (traceBuilder == null) {
            return;
        }

        double totalTime = (System.currentTimeMillis() - traceBuilder.traceTime) / 1000.0;
        appendTraceLog(traceBuilder, "TOTAL-USED", String.format("  %.3fs", totalTime));
    }

    public static void appendObjFields(XContext context, Object obj) {
        Reflect.on(obj).fields().forEach((f, r) -> {
            appendTraceLog(context, f, traceReadable(r.get()));
        });
    }

    public static void appendObjFields(TraceHelper.TraceBuilder traceBuilder, Object obj) {
        Reflect.on(obj).fields().forEach((f, r) -> {
            appendTraceLog(traceBuilder, f, traceReadable(r.get()));
        });
    }

    public static Object traceReadable(Object v) {
        if (v == null) {
            return v;
        }
        if (v instanceof String) {
            return v;
        }
        if (v instanceof Number) {
            return v;
        }
        if (v instanceof Boolean) {
            return v;
        }
        if (v instanceof Collections) {
            return XJsonUtils.toJson(v);
        }
        return v.toString();
    }

    public static void removeTraceBuilder(XContext context) {
        if (context == null) {
            return;
        }
        TraceHelper.TraceBuilder traceBuilder = getTraceBuilder(context);
        if (traceBuilder != null) {
            context.removeAdditionalData(TraceHelper.FORM_DATA_TRACE_BUILDER);
            if (traceBuilders.contains(traceBuilder.getTraceId())) {
                traceBuilders.remove(traceBuilder.getTraceId());
            }
        }
    }
}
