package pps.core.system.service;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import pps.core.common.config.TraceLogConfig;
import pps.core.common.utils.CounterBuilder;
import pps.core.common.utils.TraceHelper;
import pps.core.system.entity.BaseInfoTraceLogEnt;
import pps.core.system.mapper.BaseInfoTraceLogMapper;
import xstartup.base.XContext;
import xstartup.core.base.helper.XThreadHelper;

import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Date;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.LongAdder;

public enum TraceLogger {


    INSTANCE;

    private final LinkedBlockingQueue<TraceHelper.TraceBuilder> traceBuilderQueue =
            new LinkedBlockingQueue<TraceHelper.TraceBuilder>(10240);
    private final Thread workerThread = new Thread(new Runnable() {
        @Override
        public void run() {
            executeWork();
        }
    });

    private BaseInfoTraceLogMapper formTraceLogMapper = null;
    private final LongAdder exceptionCounter = new LongAdder();
    private final LongAdder insertedCounter = new LongAdder();

    public void start() {
        workerThread.start();
    }

    public long getExceptionCount() {
        return exceptionCounter.longValue();
    }

    public long getInsertCount() {
        return insertedCounter.longValue();
    }

    public long tasks() {
        return this.traceBuilderQueue.size();
    }


    public void add(XContext context) {
        TraceHelper.TraceBuilder traceBuilder = TraceHelper.getTraceBuilder(context);
        if (traceBuilder == null) {
            return;
        }

        add(traceBuilder, context);
    }

    public void add(TraceHelper.TraceBuilder builder) {
        this.add(builder, null);
    }

    private ConcurrentHashMap<String, String> traceIds = new ConcurrentHashMap<String, String>();

    public void add(TraceHelper.TraceBuilder builder, XContext context) {
        if (this.formTraceLogMapper == null && context != null) {
            this.formTraceLogMapper = context.getBean(BaseInfoTraceLogMapper.class);
        }

        if (builder == null || formTraceLogMapper == null) {
            return;
        }

        // 读取配置
        Boolean enableTraceLog = context == null ? true : TraceLogConfig.checkEnableTraceLog(context);// TraceLogConfig.enableTraceLog != null ? TraceLogConfig.enableTraceLog.getValue(context) : null;
        if (context != null && enableTraceLog != null && !enableTraceLog.booleanValue()) {
            return;
        }

        try {
            if( context != null) {
                context.getLogger().info(builder.buildTraceLog());
            }
            traceBuilderQueue.put(builder);
            CounterBuilder.globalCounterBuilder.addCounter("TraceLog-Add");
        } catch (InterruptedException e) {
            exceptionCounter.increment();
        }
    }

    private long preCheckTimeTicket = 0;

    private void executeWork() {
        while (true) {
            checkAndCleanData();
            try {
                TraceHelper.TraceBuilder traceBuilder = traceBuilderQueue.take();
                BaseInfoTraceLogEnt ent = new BaseInfoTraceLogEnt();
                ent.setCreateTime(new Date());
                ent.setTraceid(traceBuilder.getTraceId());
                ent.setTraceInfo(traceBuilder.buildTraceLog());
                traceBuilder.clear();
                formTraceLogMapper.insert(ent);
                insertedCounter.increment();
                CounterBuilder.globalCounterBuilder.addCounter("TraceLog-Saved");
            } catch (Exception ex) {
                exceptionCounter.increment();
                CounterBuilder.globalCounterBuilder.addCounter(ex, this.getClass());
                CounterBuilder.globalCounterBuilder.addCounter("TraceLog-WorkException");
                XThreadHelper.sleep(1000);
            }
        }
    }

    private void checkAndCleanData() {
        try {
            if (formTraceLogMapper == null) {
                return;
            }

            long span = System.currentTimeMillis() - preCheckTimeTicket;
            if (span / 1000.0 / 60 / 60 > 1) {
                //1小时
                preCheckTimeTicket = System.currentTimeMillis();
                QueryWrapper<BaseInfoTraceLogEnt> qw = new QueryWrapper<BaseInfoTraceLogEnt>();
                LocalDateTime day2Before = LocalDateTime.now().plusDays(-7);
                Date date = Date.from(day2Before.atZone(ZoneId.systemDefault()).toInstant());
                qw.lambda().lt(BaseInfoTraceLogEnt::getCreateTime, date);
                int deletedRows = formTraceLogMapper.delete(qw);
                if (deletedRows > 0) {
                    CounterBuilder.globalCounterBuilder.addCounter("TraceLog-RemoveTraceLog", deletedRows);
                }
            }
        } catch (Exception ex) {
            exceptionCounter.increment();
            CounterBuilder.globalCounterBuilder.addCounter(ex, this.getClass());
        }
    }

}
